"""
Module contains base settings classes
"""

from copy import deepcopy
from typing import Any, Dict, List

import fastjsonschema


class DeepCopyMixin:  # pylint: disable-msg=too-few-public-methods
    """Deep copy mixin"""

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for attribute, value in self.__dict__.items():
            setattr(result, attribute, deepcopy(value, memo))
        return result


class BaseSettingDefinition(DeepCopyMixin):
    """
    Base setting definition class
    Attributes:
        services(list): list of services which use this setting
        schema(dict): validation schema for setting
        default(any): default setting value
        description(str): setting description
        groups(list): list of groups which this setting applies to
        external: mark that settings is external
    """

    def __init__(  # pylint: disable-msg=too-many-positional-arguments
        self, settingName: str, services: List[str], schema: dict, default: Any, description: str, groups: List[str]
    ):
        self._settingName = settingName
        self.services = services
        self.schema = schema
        self.default = default
        self.description = description
        self.groups = groups
        self.external = False
        self._validateSettings()

    def asDict(self) -> Dict[str, Any]:
        """
        Get setting definition as dict
        Returns:
            dict with setting definition
        """
        return {
            "services": self.services,
            "schema": self.schema,
            "default": self.default,
            "description": self.description,
            "groups": self.groups,
            "external": self.external,
        }

    def _validateSettings(self):
        """
        Validate settings
        Raises:
            ValueError if one settings structure does not match requirement
                    or if default values in setting does not match schema
        """

        def baseError(errorMsg):
            return ValueError(f"Setting {self._settingName}: {errorMsg}")

        if not isinstance(self.services, list):
            raise baseError("services must be list")
        if not isinstance(self.schema, dict):
            raise baseError("schema must be dict")
        if not isinstance(self.description, str):
            raise baseError("description must be string")
        if not isinstance(self.groups, list):
            raise baseError("groups must by list")
        try:
            compiledSchema = fastjsonschema.compile(self.schema)
            compiledSchema(self.default)
        except fastjsonschema.JsonSchemaException as exc:
            raise baseError(exc.message) from exc  # pylint: disable-msg=no-member

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self._settingName}>"


class ExternalSettingDefinition(DeepCopyMixin):
    """
    Base setting definition class
    Attributes:
        services(list): list of services which use this setting
        schema(dict): validation schema for setting
        groups(list): list of groups which this setting applies to
        external: mark that settings is external
    """

    def __init__(  # pylint: disable-msg=too-many-positional-arguments
        self, settingName: str, services: List[str], schema: dict, default: Any, groups: List[str]
    ):
        self._settingName = settingName
        self.services = services
        self.schema = schema
        self.groups = groups
        self.external = True
        self.default = default
        self._validateSettings()

    def asDict(self) -> Dict[str, Any]:
        """
        Get setting definition as dict
        Returns:
            dict with setting definition
        """
        return {
            "services": self.services,
            "schema": self.schema,
            "groups": self.groups,
            "external": self.external,
        }

    def _validateSettings(self):
        """
        Validate settings
        Raises:
            ValueError if one settings structure does not match requirement
                    or if default values in setting does not match schema
        """

        def baseError(errorMsg):
            return ValueError(f"Setting {self._settingName}: {errorMsg}")

        if not isinstance(self.services, list):
            raise baseError("services must be list")
        if not isinstance(self.schema, dict):
            raise baseError("schema must be dict")
        if not isinstance(self.groups, list):
            raise baseError("groups must by list")
        try:
            compiledSchema = fastjsonschema.compile(self.schema)
            compiledSchema(self.default)
        except fastjsonschema.JsonSchemaException as exc:
            raise baseError(exc.message) from exc  # pylint: disable-msg=no-member

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self._settingName}>"
